/*
 *
 *  Copyright Peter Boyle and Glasgow University 2000. 
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include "processor.h"
#include <string.h>

extern struct processor *PROC;
extern struct inst_queue_entry *HEAD;
void stall_inst_print(struct instruction *i);

void arg_sprint(struct instruction *i, int j,char *buf);
void inst_print_unknown(struct instruction *i);

/*{... OUTPUT routines*/
int human_readable = 0;
void set_human_readable (int h)
{
  human_readable = h;
}

static char *label_prefix;
static char *label_fmt = "_%s_lab%d";
void set_label_fmt(char *fmt)
{
  label_fmt = fmt;
}
void set_label_prefix(char *pfx)
{ 
  label_prefix=pfx;
}
char *make_label(int i)
{
  static char buf[84];
  if (label_prefix) 
    sprintf(buf,label_fmt,label_prefix,i);
  else 
    sprintf(buf,".lab%d",i);
    
  return buf;
}


char *make_reg_name (int kind, int reg )
{
  char * str = (char *)malloc(6);
  switch(kind){
    case Iregs:  
      sprintf(str,PROC->iregs->fmt,reg);
      break;
    case Fregs:
      switch ( PROC->id ){
      case USPARCII: /* Ultrasparc nastily uses 2 single fregs as one double*/
                     /* Bleuchh.... so dfreg 31 is %f62 etc.. */
	sprintf(str,PROC->fregs->fmt,2*reg);
        break;
      default:
	sprintf(str,PROC->fregs->fmt,reg);
        break;
      }
      break;
    default:
      puts("Error: Bad register kind");
      exit(0);
      break;
  }
  return(str);
}

void dump_instruction_queue(void)
{
  struct inst_queue_entry *inst;
  struct offset *off;
  int handle;
  int i;
  int pipe;
  inst = HEAD;

  handle = 0;
  printf("/* \n");
  printf(" * BAGEL machine generated output.  \n");
  printf(" * Copyright UKQCD Collaboration  \n");
  printf(" * This software is provided for NON-COMMERCIAL use only,  \n");
  printf(" * and may not be redistributed without permission.  \n");
  printf(" * It is provided as is, and is not guaranteed fit for any purpose.\n");
  printf(" * Written by Peter Boyle\n");
  printf(" */  \n");

  if (PROC->id >= UNKNOWN){
    human_readable = 1;
    while ( (off= get_offset_struct (handle++))!=NULL ) {
      printf("#define %s %d \n",off->name, off->offst);
    }
    printf("\n");
    while (inst !=NULL){
      inst_print_unknown(inst->inst);
      inst = inst->next;
    }
    return;

  } else {
    if ( human_readable ){
      while ( (off= get_offset_struct (handle++))!=NULL ) {
        printf("#define %s %d \n",off->name, off->offst);
      }
      printf("\n");
      dump_regs(Fregs);
      dump_regs(Iregs);
    }
    while (inst !=NULL){
      inst_print(inst->inst);
      if ( human_readable && inst->inst->staller ){
        stall_inst_print(inst->inst->staller);
      }
      inst = inst->next;
    }
  }
  fflush(stdout);
}

void arg_print(struct instruction *i, int j)
{
  int atype=i->argtypes[j];

  if ( human_readable ) {
    switch( atype ) {
    case Fregs:
      if ( PROC->fregs->allocated[i->args[j]] ){
        printf("%s",PROC->fregs->mnemonic[i->args[j]]);
      }
      else{
        printf("%s",make_reg_name(Fregs,i->args[j]));
      }
      break;

    case Iregs:
      if ( PROC->iregs->allocated[i->args[j]] ){
        printf("%s",PROC->iregs->mnemonic[i->args[j]]);
      }
      else{
        printf("%s",make_reg_name(Iregs,i->args[j]));
      }
      break;
    case SIMM:
      printf("%s",get_offset_name(i->args[j]));break;
    case Label:
      printf(make_label(i->args[j]));break;
    default: 
      printf(" Bad instruction\n ");break;
    }
  }
  else {
      switch( atype ) {
      case Fregs:
        printf("%s",make_reg_name(Fregs,i->args[j]));break;
      case Iregs:
        printf("%s",make_reg_name(Iregs,i->args[j]));break;
      case SIMM:
        printf("%d",get_offset(i->args[j]));break;
      case Label:
        printf(make_label(i->args[j]));break;
      default: 
        printf(" Bad instruction\n ");break;
      }
  }

}

void inst_print(struct instruction *i)
{
  int j,p;


  /*
   * We've taken the knowledge of effective addressing out of the arg types to
   * make dependency checking logic simpler.
   * Consequently we need to check here whether the opcode demands construction
   * of n(reg) or n,reg arguments
   */

  if ( i->code == Enter_Routine && i->pipe == DIRECTIVE) { 
    set_label_prefix(i->mnemonic);
  }

  switch ( i-> pipe ){
  case DIRECTIVE:
    directive_print(i); /*These are so architecture specific I give up generality*/
    break;
  case CACHPIPE:
    cache_print ( i); /*These are so architecture specific I give up generality*/
    break;
  case LOADPIPE:
    if ( i->code == LOAD_IMM ) {
      printf("\t%s ",i->mnemonic);
      if ( PROC->switchargs[i->pipe] ){
         arg_print(i,1);
         printf(",");
         arg_print(i,0);
      }
      else{
         arg_print(i,0);
         printf(",");
         arg_print(i,1);
      }
      printf("\n");
      break;
    } else {
      print_eff_addr(i);
      printf("\n");
      break;
    }
  
  case FLODPIPE:
    if ( (i->code == CLOAD_INDEXED) || (i->code ==CS_LOAD_INDEXED) ) {
      printf("\t%s ",i->mnemonic);
      arg_print(i,0);
      printf(",");
      arg_print(i,1);
      printf(",");
      arg_print(i,2);
    } else { 
      print_eff_addr(i);
    }
    printf("\n");
    break;
  case FSTRPIPE: /*Generate code for effective addresses*/
    if ( ( i->code == CSAVE_INDEXED ) || ( i->code == CS_SAVE_INDEXED ) ){
      printf("\t%s ",i->mnemonic);
      arg_print(i,0);
      printf(",");
      arg_print(i,1);
      printf(",");
      arg_print(i,2);
    } else { 
      print_eff_addr(i);
    }
    printf("\n");
    break;

  case STORPIPE:
    print_eff_addr(i);
    printf("\n");
    break;

  case BRCHPIPE:
    printf("\t%s ",i->mnemonic);
    if ( PROC->switchargs[i->pipe] ){
      for(j=1;j<i->narg;j++) {
       arg_print(i,j);
       printf(" , ");
      }
      arg_print(i,0);
    }
    else {
      arg_print(i,0);
      if ( i->narg > 1 ) 
        printf(" , ");
      for(j=1;j<i->narg;j++) {
        arg_print(i,j);
        if ( j<i->narg-1 ) printf(" , ");
      }
    }
    if ( PROC->delayslot == 1 ) /*Insert a nop in branch delay slot */
      printf("\n\t%s",PROC->iset->instructions[DIRECTIVE][NOP].mnemonic);            
    printf("\n");
    break;

  default: 

    if ( human_readable ){
      for(j=0;j<PROC->npipes;j++) 
        if ( PROC->pipes[j].issuegroups & ( 1<<i->pipe ) ) p = j ;
      if ( p == 0 ) printf("\n");
      printf("\t");
      /*      for(j=0;j<p;j++)
	      printf("   ");*/
    } 
    else printf("\t");

    printf("%s ",i->mnemonic);

    if ( PROC->switchargs[i->pipe] ){
      for(j=1;j<i->narg;j++) {
        arg_print(i,j);
        printf(" , ");
      }
      arg_print(i,0);
    }
    else {
      arg_print(i,0);
      if ( i->narg > 1 ) 
        printf(" , ");
      for(j=1;j<i->narg;j++) {
        arg_print(i,j);
        if ( j<i->narg-1 ) printf(" , ");
      }
    }
    
    printf("\n");
    break;
  }
  return;
}




void stall_inst_print(struct instruction *i)
{
  int j,p;


  /*
   * We've taken the knowledge of effective addressing out of the arg types to
   * make dependency checking logic simpler.
   * Consequently we need to check here whether the opcode demands construction
   * of n(reg) or n,reg arguments
   */
  if ( i->pipe == DIRECTIVE || i->pipe == CACHPIPE ) return;

  printf("\t\t\t/* Last Dependency:");
  switch ( i-> pipe ){
  case DIRECTIVE:
    break;
  case CACHPIPE:
    cache_print ( i);
    break;
  case LOADPIPE:
    if ( i->code == LOAD_IMM ) {
      printf("\t%s ",i->mnemonic);
      if ( PROC->switchargs[i->pipe] ){
         arg_print(i,1);
         printf(",");
         arg_print(i,0);
      }
      else{
         arg_print(i,0);
         printf(",");
         arg_print(i,1);
      }
      break;
    } else {
      print_eff_addr(i);
      break;
    }
  case STORPIPE:
  case FLODPIPE:
  case FSTRPIPE: /*Generate code for effective addresses*/
    print_eff_addr(i);
    break;

  case BRCHPIPE:
    printf("\t%s ",i->mnemonic);
    if ( PROC->switchargs[i->pipe] ){
      for(j=1;j<i->narg;j++) {
       arg_print(i,j);
       printf(" , ");
      }
      arg_print(i,0);
    }
    else {
      arg_print(i,0);
      if ( i->narg > 1 ) 
        printf(" , ");
      for(j=1;j<i->narg;j++) {
        arg_print(i,j);
        if ( j<i->narg-1 ) printf(" , ");
      }
    }
    if ( PROC->delayslot == 1 ) /*Insert a nop in branch delay slot */
      printf("\n\t%s",PROC->iset->instructions[DIRECTIVE][NOP].mnemonic);            
    break;

  default: 

    printf("\t");
    printf("%s ",i->mnemonic);

    if ( PROC->switchargs[i->pipe] ){
      for(j=1;j<i->narg;j++) {
        arg_print(i,j);
        printf(" , ");
      }
      arg_print(i,0);
    }
    else {
      arg_print(i,0);
      if ( i->narg > 1 ) 
        printf(" , ");
      for(j=1;j<i->narg;j++) {
        arg_print(i,j);
        if ( j<i->narg-1 ) printf(" , ");
      }
    }
    break;
  }
  printf("*/\n");
  return;
}



void inst_print_unknown(struct instruction *i)
{
  int j,p;
  char arg0[80];
  char arg1[80];
  char arg2[80];
  char arg3[80];
  /*
   * We've taken the knowledge of effective addressing out of the arg types to
   * make dependency checking logic simpler.
   * Consequently we need to check here whether the opcode demands construction
   * of n(reg) or n,reg arguments
   */
  if ( i->code == Enter_Routine && i->pipe == DIRECTIVE) { 
    set_label_prefix(i->mnemonic);
  }

  switch ( i-> pipe ){
  case DIRECTIVE:
    directive_print_unknown(i); 
    break;
  case CACHPIPE:
    arg_sprint(i,0,arg0);
    arg_sprint(i,1,arg1);
    printf("/*preload %s %s */\n",arg0,arg1);
    break;
  case LOADPIPE:
    if ( i->code == LOAD_IMM ) {
      arg_print(i,0);
      printf("=");
      arg_print(i,1);
      printf(";\n");
    }
    else{
      arg_sprint(i,0,arg0);
      arg_sprint(i,1,arg1);
      arg_sprint(i,2,arg2);
      printf(i->mnemonic,arg0,arg1,arg2);
      printf("\n");
    }
    break;
  case PRAGMA:
    printf("/*%s %d*/\n",i->mnemonic,i->args[0]);
    break;
  case FLODPIPE:
      arg_sprint(i,0,arg0);
      arg_sprint(i,1,arg1);
      arg_sprint(i,2,arg2);
      printf(i->mnemonic,arg0,arg1,arg2);
      printf("\n");
      break;
  case STORPIPE:
  case FSTRPIPE: /*Generate code for effective addresses*/
      arg_sprint(i,0,arg0);
      arg_sprint(i,1,arg1);
      arg_sprint(i,2,arg2);
      printf(i->mnemonic,arg1,arg2,arg0);
      printf("\n");
    break;

  case BRCHPIPE:
    switch( i->code ){
    case BRANCH_GT:
      printf("  if ( (long) ");
      arg_print(i,0);
      printf(" >  ");
      printf("0 ) goto ");
      arg_print(i,1);
      printf(";\n");
      break;
    case BRANCH_EQ:
      printf("  if ( ");
      arg_print(i,0);
      printf(" ==  ");
      printf("0 ) goto ");
      arg_print(i,1);
      printf(";\n");
      break;
    case BRANCH_LE:
      printf("  if ( (long) ");
      arg_print(i,0);
      printf(" <= ");
      printf("0 ) goto ");
      arg_print(i,1);
      printf(";\n");
      break;
    case BRANCH_LT:
      printf("  if (  (long) ");
      arg_print(i,0);
      printf(" < ");
      printf("0 ) goto ");
      arg_print(i,1);
      printf(";\n");
      break;
    case BRANCH_GE:
      printf("  if (  (long) ");
      arg_print(i,0);
      printf(" >= ");
      printf("0 ) goto ");
      arg_print(i,1);
      printf(";\n");
      break;
    case BRANCH:
      printf(" goto ");
      arg_print(i,0);
      printf(";\n");
      break;
    case BRANCH_RET:    
      printf("\nreturn;\n");
      break;
    }
    break;

  default: 

    switch(i->narg){
    case 1:
      arg_sprint(i,0,arg0);
      printf(i->mnemonic,arg0);
      printf("\n");
      break;
    case 2:
      arg_sprint(i,0,arg0);
      arg_sprint(i,1,arg1);
      printf(i->mnemonic,arg0,arg1);
      printf("\n");
      break;
    case 3:
      arg_sprint(i,0,arg0);
      arg_sprint(i,1,arg1);
      arg_sprint(i,2,arg2);
      printf(i->mnemonic,arg0,arg1,arg2);
      printf("\n");
      break;
    case 4:
      arg_sprint(i,0,arg0);
      arg_sprint(i,1,arg1);
      arg_sprint(i,2,arg2);
      arg_sprint(i,3,arg3);
      printf(i->mnemonic,arg0,arg1,arg2,arg3);
      printf("\n");
      break;
    case 0:
      printf(i->mnemonic);
      printf("\n");
      break;
    default:
      printf("Unkown inst format!!!");
      break;
    }
    break;
  }
  return;
}

void dump_regs( int kind)
{
  struct regs *set;
  int num;

  test_proc();
  switch ( kind ){
    case Fregs: 
      set = PROC->fregs;
      break;
    case Iregs: 
      set = PROC->iregs;
      break;
    default: puts("Bad Arugment: Free reg"); exit(0);
  }
  for ( num = 0 ; num < set->regfile;num++){

    if ( set->allocated[num] ){
      printf( "#define %s %s\n",set->mnemonic[num],make_reg_name(kind,num));
    }
  }
  return;
}


void print_eff_addr(struct instruction *i)
{
  int j,p;

  if ( human_readable) {
    for(j=0;j<PROC->npipes;j++) 
      if ( PROC->pipes[j].issuegroups & ( 1<<i->pipe ) ) p = j ;
    if ( p == 0 ) printf("/*New cycle*/\n");
    printf("\t");
    for(j=0;j<p;j++)
      printf("   ");
  } else printf("\t");

  

  switch ( PROC->id ){
  case USPARCII:
  case USPARCII_SINGLE:
    printf("%s ",i->mnemonic);
    if ( PROC->switchargs[i->pipe] ){
       printf("[");
       arg_print(i,2);
       printf("+");
       arg_print(i,1);
       printf("],");
       arg_print(i,0);
    }
    else{
       arg_print(i,0);
       printf(",[");
       arg_print(i,2);
       printf("+");
       arg_print(i,1);
       printf("]");
    }
    break;
  default:
    printf("%s ",i->mnemonic);
    if ( PROC->switchargs[i->pipe] ){
       arg_print(i,1);
       printf("(");
       arg_print(i,2);
       printf("),");
       arg_print(i,0);
    }
    else{
       arg_print(i,0);
       printf(",\t");
       arg_print(i,1);
       printf("(");
       arg_print(i,2);
       printf(")");
    }
    break;
  }

}

void arg_sprint(struct instruction *i, int j,char *buf)
{
  int atype=i->argtypes[j];

  if ( human_readable ) {
    switch( atype ) {
    case Fregs:
      if ( PROC->fregs->allocated[i->args[j]] ){
        sprintf(buf,"%s",PROC->fregs->mnemonic[i->args[j]]);
      }
      else{
        sprintf(buf,"%s",make_reg_name(Fregs,i->args[j]));
      }
      break;

    case Iregs:
      if ( PROC->iregs->allocated[i->args[j]] ){
        sprintf(buf,"%s",PROC->iregs->mnemonic[i->args[j]]);
      }
      else{
        sprintf(buf,"%s",make_reg_name(Iregs,i->args[j]));
      }
      break;
    case SIMM:
      sprintf(buf,"%s",get_offset_name(i->args[j]));break;
    case Label:
      sprintf(buf,make_label(i->args[j]));break;
    default: 
      printf(" Bad instruction\n ");break;
    }
  }
  else {
      switch( atype ) {
      case Fregs:
        sprintf(buf,"%s",make_reg_name(Fregs,i->args[j]));break;
      case Iregs:
        sprintf(buf,"%s",make_reg_name(Iregs,i->args[j]));break;
      case SIMM:
        sprintf(buf,"%d",get_offset(i->args[j]));break;
      case Label:
        sprintf(buf,make_label(i->args[j]));break;
      default: 
        printf(" Bad instruction\n ");break;
      }
  }

}



/* ...END OUTPUT routines}*/
